"""
FortiManager Class
"""

import logging
import re
from pathlib import Path
from time import sleep
from typing import Any, Dict, List, Optional

import requests

from .fortinet import Fortinet

log = logging.getLogger("fotoobo")


class FortiManager(Fortinet):  # pylint: disable=too-many-public-methods
    """
    Represents one FortiManager (digital twin)
    """

    def __del__(self) -> None:
        """The destructor"""
        if self.session_key and not self.session_path:
            self.logout()

    def __init__(self, hostname: str, username: str, password: str, **kwargs: Any) -> None:
        """
        Set some initial parameters.

        Args:
            hostname: The hostname of the FortiGate to connect to
            username: The username
            password: The password

        Keyword Args:
            session_dir: The path where to load/save the FortiManager session key
            **kwargs:    See Fortinet class for more available arguments
        """
        super().__init__(hostname, **kwargs)
        self.api_url = f"https://{self.hostname}:{self.https_port}/jsonrpc"
        self.password = password
        self.username = username
        self.session_key: str = ""
        self.session_path: str = kwargs.get("session_path", "")
        self.type = "fortimanager"
        self.ignored_adoms = [
            "FortiAnalyzer",
            "FortiAuthenticator",
            "FortiCache",
            "FortiCarrier",
            "FortiClient",
            "FortiDDoS",
            "FortiDeceptor",
            "FortiFirewall",
            "FortiFirewallCarrier",
            "FortiMail",
            "FortiManager",
            "FortiNAC",
            "FortiProxy",
            "FortiSandbox",
            "FortiWeb",
            "Syslog",
            "Unmanaged_Devices",
            "others",
            "root",
            "rootp",
        ]

    def api_delete(self, url: str) -> requests.models.Response:
        """DELETE method for API requests

        Args:
            url: API endpoint to access

        Result:
            FortiManager result item
        """
        payload = {
            "method": "delete",
            "params": [{"url": f"{url}"}],
        }
        return self.api("post", payload=payload)

    def api_get(
        self, url: str, params: Optional[dict[str, Any]] = None, timeout: Optional[float] = None
    ) -> requests.models.Response:
        """GET method for API requests

        Args:
            url:     API endpoint to access
            params:  Additional query parameters if needed
            timeout: The requests read timeout in seconds

        Result:
            FortiManager result item
        """
        _params = {"url": f"{url}"}
        if params:
            _params = {**_params, **params}

        payload = {
            "method": "get",
            "params": [_params],
        }
        return self.api("post", payload=payload, timeout=timeout)

    def api(  # pylint: disable=too-many-arguments
        self,
        method: str,
        url: str = "",
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        payload: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
    ) -> requests.models.Response:
        """
        API request to a FortiManager device.

        It uses the super.api method but it has to enrich the payload in post requests with the
        needed session key.

        Args:
            method:  Request method from [get, post]
            url:     Rest API URL to request data from
            headers: Dictionary with headers (if needed)
            params:  Dictionary with parameters (if needed)
            payload: JSON body for post requests (if needed)
            timeout: The requests read timeout in seconds

        Returns:
            Response from the request
        """
        if not self.session_key:
            self.login()

        payload = payload or {}
        if method.lower() == "post":
            payload["session"] = self.session_key

        return super().api(
            method, url, headers=headers, payload=payload, params=params, timeout=timeout
        )

    def assign_all_objects(self, adoms: str, policy: str) -> int:
        """
        Copies all objects from the global ADOM database to a given ADOM.

        Args:
            adoms:  The ADOM to assign the global policy/objects to. If you specify an invalid ADOM
                    name you'll get a permission error. You can specify multiple ADOMs by separating
                    them with a comma (no spaces)
            policy: Specify the global policy to assign [Default: 'default'].

        Returns:
            The id of the FortiManager task created or 0 (zero) if unsuccessful

        """
        task_id = 0
        adom_payload = []

        for adom in adoms.split(","):
            adom_payload.append({"adom": adom, "excluded": "disable"})

        payload = {
            "method": "exec",
            "params": [
                {
                    "data": {
                        "flags": ["cp_all_objs"],
                        "pkg": policy,
                        "target": adom_payload,
                    },
                    "url": "/securityconsole/assign/package",
                }
            ],
        }
        response = self.api("post", payload=payload)

        if response.status_code == 200:
            if response.json()["result"][0]["status"]["code"] == 0:
                task_id = response.json()["result"][0]["data"]["task"]
                log.debug("Assign task created with id '%s'", task_id)

            else:
                log.debug(
                    "Did not assign to '%s' with error '%s'",
                    adoms,
                    response.json()["result"][0]["status"]["message"],
                )
                task_id = 0

        return task_id

    def delete_adom_address(self, adom: str, address: str, dry: bool = False) -> dict[str, Any]:
        """
        Delete an address from an ADOM in FortiManager

        Args:
            adom:    The ADOM to delete the address from
            address: The address to delete
            dry:     Set to True to enable dry-run (no changes on FortiManager)

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}
        if not dry:
            url: str = f"/pm/config/adom/{adom}/obj/firewall/address/{address}"
            result = self.api_delete(url).json()["result"][0]

        else:
            log.info("DRY-RUN: Would remove address '%s' in ADOM '%s'", address, adom)

        return result

    def delete_adom_address_group(self, adom: str, group: str, dry: bool = False) -> dict[str, Any]:
        """
        Delete an address group from an ADOM in FortiManager

        Args:
            adom:  The ADOM to delete the address group from
            group: The address group to delete
            dry:   Set to True to enable dry-run (no changes on FortiManager)

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}
        if not dry:
            url: str = f"/pm/config/adom/{adom}/obj/firewall/addrgrp/{group}"
            result = self.api_delete(url).json()["result"][0]

        else:
            log.info("DRY-RUN: Would remove address group '%s' in ADOM '%s'", group, adom)

        return result

    def delete_adom_service(self, adom: str, service: str, dry: bool = False) -> dict[str, Any]:
        """
        Delete a service from an ADOM in FortiManager

        Args:
            adom:    The ADOM to delete the address from
            service: The service to delete
            dry:     Set to True to enable dry-run (no changes on FortiManager)

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}
        if not dry:
            url: str = f"/pm/config/adom/{adom}/obj/firewall/service/custom/{service}"
            result = self.api_delete(url).json()["result"][0]

        else:
            log.info("DRY-RUN: Would remove service '%s' in ADOM '%s'", service, adom)

        return result

    def delete_adom_service_group(self, adom: str, group: str, dry: bool = False) -> dict[str, Any]:
        """
        Delete a service group from an ADOM in FortiManager

        Args:
            adom:  The ADOM to delete the address group from
            group: The address group to delete
            dry:   Set to True to enable dry-run (no changes on FortiManager)

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}
        if not dry:
            url: str = f"/pm/config/adom/{adom}/obj/firewall/service/group/{group}"
            result = self.api_delete(url).json()["result"][0]

        else:
            log.info("DRY-RUN: Would remove service group '%s' in ADOM '%s'", group, adom)

        return result

    def delete_global_address(self, address: str, dry: bool = False) -> dict[str, Any]:
        """
        Delete a global address from FortiManager

        To be sure to not delete used objects we configure the FortiManager as follows:

            config system admin setting
                set objects-force-deletion disable

        Before deleting a global address object we have to check if it's in use in any ADOM. Only
        after we found that no ADOM uses the object we can safely delete it.

        Fortinet documentation links:\n
        https://docs.fortinet.com/document/fortimanager/7.0.12/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.2.7/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.4.3/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.6.0/administration-guide/322714

        Args:
            address: The global address to delete
            dry:     Set to True to enable dry-run (no changes on FortiManager)

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}

        # Get the address object with 'scope member' information
        address_object = self.get_global_address(address, scope_member=True)
        if address_object["status"]["code"] == 0:
            # Generate a list of ADOMs where the object is used. Therefore we get the address object
            # from FortiManager with the 'scope member' option. If the object is used in any ADOM it
            # is listed in the key 'scope member'. If the object is not used in any ADOM the
            # 'scope member' key is not present.
            if "scope member" in address_object["data"]:
                used_adoms = [_["name"] for _ in address_object["data"]["scope member"]]
                log.debug("'%s' is used in ADOM '%s'", address, ",".join(used_adoms))

            else:
                used_adoms = []

            if not dry:
                # Try to delete the address group object in every ADOM
                blocked_adoms = []
                for adom in used_adoms:
                    if self.delete_adom_address(adom, address)["status"]["code"] not in [
                        -3,
                        0,
                    ]:
                        blocked_adoms.append(adom)

                if blocked_adoms:
                    log.warning("'%s' blocked by ADOM '%s'", address, ",".join(blocked_adoms))
                    result = address_object
                    result["status"] = {
                        "code": 601,
                        "message": f"Used in ADOM {','.join(blocked_adoms)}",
                    }

                else:
                    # Try to delete the global address object
                    url: str = f"/pm/config/global/obj/firewall/address/{address}"
                    result = self.api_delete(url).json()["result"][0]

            else:
                log.info("DRY-RUN: Would remove global address '%s'", address)

        else:
            result = address_object

        return result

    def delete_global_address_group(self, group: str, dry: bool = False) -> dict[str, Any]:
        """
        Delete a global address group from FortiManager

        To be sure to not delete used objects we configure the FortiManager as follows:

            config system admin setting
                set objects-force-deletion disable

        Before deleting a global address group object we have to check if it's in use in any ADOM.
        Only after we found that no ADOM uses the object we can safely delete it.

        Fortinet documentation links:\n
        https://docs.fortinet.com/document/fortimanager/7.0.12/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.2.7/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.4.3/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.6.0/administration-guide/322714

        Args:
            group: The global address group to delete
            dry:   Set to True to enable dry-run (no changes on FortiManager)

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}

        # Get the address group object with 'scope member' information
        address_group_object = self.get_global_address_group(group, scope_member=True)
        if address_group_object["status"]["code"] == 0:

            # Generate a list of ADOMs where the object is used. Therefore we get the address group
            # object from FortiManager with the 'scope member' option. If the object is used in any
            # ADOM it is listed in the key 'scope member'. If the object is not used in any ADOM the
            # 'scope member' key is not present.
            if "scope member" in address_group_object["data"]:
                used_adoms = [_["name"] for _ in address_group_object["data"]["scope member"]]
                log.debug("'%s' is used in ADOM '%s'", group, ",".join(used_adoms))

            else:
                used_adoms = []

            if not dry:
                # Try to delete the address group object in every ADOM
                blocked_adoms = []
                for adom in used_adoms:
                    if not self.delete_adom_address_group(adom, group)["status"]["code"] in [-3, 0]:
                        blocked_adoms.append(adom)

                if blocked_adoms:
                    log.warning("'%s' blocked by ADOM '%s'", group, ",".join(blocked_adoms))
                    result = address_group_object
                    result["status"] = {
                        "code": 601,
                        "message": f"Used in ADOM {','.join(blocked_adoms)}",
                    }

                else:
                    # Try to delete the global address group object
                    url: str = f"/pm/config/global/obj/firewall/addrgrp/{group}"
                    result = self.api_delete(url).json()["result"][0]

            else:
                log.info("DRY-RUN: Would remove global address group '%s'", group)

        else:
            result = address_group_object

        return result

    def delete_global_service(self, service: str, dry: bool = False) -> dict[str, Any]:
        """
        Delete a global service from FortiManager

        To be sure to not delete used objects we configure the FortiManager as follows:

            config system admin setting
                set objects-force-deletion disable

        Before deleting a global service object we have to check if it's in use in any ADOM. Only
        after we found that no ADOM uses the object we can safely delete it.

        Fortinet documentation links:\n
        https://docs.fortinet.com/document/fortimanager/7.0.12/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.2.7/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.4.3/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.6.0/administration-guide/322714

        Args:
            service: The global service to delete
            dry:     Set to True to enable dry-run (no changes on FortiManager)

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}

        # Get the service object with 'scope member' information
        service_object = self.get_global_service(service, scope_member=True)
        if service_object["status"]["code"] == 0:

            # Generate a list of ADOMs where the object is used. Therefore we get the service object
            # from FortiManager with the 'scope member' option. If the object is used in any ADOM it
            # is listed in the key 'scope member'. If the object is not used in any ADOM the
            # 'scope member' key is not present.
            if "scope member" in service_object["data"]:
                used_adoms = [_["name"] for _ in service_object["data"]["scope member"]]
                log.debug("'%s' is used in ADOM '%s'", service, ",".join(used_adoms))

            else:
                used_adoms = []

            if not dry:
                # Try to delete the address group  object in every ADOM
                blocked_adoms = []
                for adom in used_adoms:
                    if not self.delete_adom_service(adom, service)["status"]["code"] in [-3, 0]:
                        blocked_adoms.append(adom)

                if blocked_adoms:
                    log.warning("'%s' blocked by ADOM '%s'", service, ",".join(blocked_adoms))
                    result = service_object
                    result["status"] = {
                        "code": 601,
                        "message": f"Used in ADOM {','.join(blocked_adoms)}",
                    }

                else:
                    # Try to delete the global service object
                    url: str = f"/pm/config/global/obj/firewall/service/custom/{service}"
                    result = self.api_delete(url).json()["result"][0]

            else:
                log.info("DRY-RUN: Would remove global service '%s'", service)

        else:
            result = service_object

        return result

    def delete_global_service_group(self, group: str, dry: bool = False) -> dict[str, Any]:
        """
        Delete a global service group from FortiManager

        To be sure to not delete used objects we configure the FortiManager as follows:

            config system admin setting
                set objects-force-deletion disable

        Before deleting a global service group object we have to check if it's in use in any ADOM.
        Only after we found that no ADOM uses the object we can safely delete it.

        Fortinet documentation links:\n
        https://docs.fortinet.com/document/fortimanager/7.0.12/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.2.7/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.4.3/administration-guide/322714
        https://docs.fortinet.com/document/fortimanager/7.6.0/administration-guide/322714

        Args:
            group: The global service group to delete
            dry:   Set to True to enable dry-run (no changes on FortiManager)

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}

        # Get the service group object with 'scope member' information
        service_group_object = self.get_global_service_group(group, scope_member=True)
        if service_group_object["status"]["code"] == 0:

            # Generate a list of ADOMs where the object is used. Therefore we get the service group
            # object from FortiManager with the 'scope member' option. If the object is used in any
            # ADOM it is listed in the key 'scope member'. If the object is not used in any ADOM the
            # 'scope member' key is not present.
            if "scope member" in service_group_object["data"]:
                used_adoms = [_["name"] for _ in service_group_object["data"]["scope member"]]
                log.debug("'%s' is used in ADOM '%s'", group, ",".join(used_adoms))

            else:
                used_adoms = []

            if not dry:
                # Try to delete the service group object in every ADOM
                blocked_adoms = []
                for adom in used_adoms:
                    if not self.delete_adom_service_group(adom, group)["status"]["code"] in [-3, 0]:
                        blocked_adoms.append(adom)

                if blocked_adoms:
                    log.warning("'%s' blocked by ADOM '%s'", group, ",".join(blocked_adoms))
                    result = service_group_object
                    result["status"] = {
                        "code": 601,
                        "message": f"Used in ADOM {','.join(blocked_adoms)}",
                    }

                else:
                    # Try to delete the global service group object
                    url: str = f"/pm/config/global/obj/firewall/service/group/{group}"
                    result = self.api_delete(url).json()["result"][0]

            else:
                log.info("DRY-RUN: Would remove global service group '%s'", group)

        else:
            result = service_group_object

        return result

    def get_adoms(self, ignored_adoms: Optional[List[str]] = None) -> list[Any]:
        """
        Get FortiManager ADOM list

        Some of the ADOMs are ignored by default as the are not used in most cases

        Returns:
            List of FortiManager ADOMs
        """
        if not ignored_adoms:
            ignored_adoms = self.ignored_adoms

        fmg_adoms: List[Any] = []
        payload = {"method": "get", "params": [{"url": "/dvmdb/adom"}]}
        response = self.api("post", payload=payload)

        for adom in response.json()["result"][0]["data"]:
            if not adom["name"] in ignored_adoms:
                fmg_adoms.append(adom)

        return fmg_adoms

    def get_global_address(self, address: str, scope_member: bool = False) -> dict[str, Any]:
        """
        Get an address object from global ADOM

        Args:
            address:      The address to get
            scope_member: Whether the scope member attribute should be included in the response

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}
        url: str = f"/pm/config/global/obj/firewall/address/{address}"

        if scope_member:
            params = {"option": ["scope member"]}
            response = self.api_get(url, params)

        else:
            response = self.api_get(url)

        result = response.json()["result"][0]

        return result

    def get_global_addresses(self) -> dict[str, Any]:
        """
        Get the global address database

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = self.api_get(
            "/pm/config/global/obj/firewall/address",
            timeout=10,
        ).json()["result"][0]

        return result

    def get_global_address_group(self, group: str, scope_member: bool = False) -> dict[str, Any]:
        """
        Get an address group object from the global ADOM

        Args:
            group:        The address group to get
            scope_member: Whether the scope member attribute should be included in the response

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}
        url: str = f"/pm/config/global/obj/firewall/addrgrp/{group}"

        if scope_member:
            params = {"option": ["scope member"]}
            response = self.api_get(url, params)

        else:
            response = self.api_get(url)

        result = response.json()["result"][0]

        return result

    def get_global_address_groups(self) -> dict[str, Any]:
        """
        Get the global address group database

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = self.api_get(
            "/pm/config/global/obj/firewall/addrgrp",
            timeout=10,
        ).json()["result"][0]

        return result

    def get_global_service(self, service: str, scope_member: bool = False) -> dict[str, Any]:
        """
        Get a service object from global ADOM

        Args:
            service:      The service to get
            scope_member: Whether the scope member attribute should be included in the response

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}
        url: str = f"/pm/config/global/obj/firewall/service/custom/{service}"

        if scope_member:
            params = {"option": ["scope member"]}
            response = self.api_get(url, params)

        else:
            response = self.api_get(url)

        result = response.json()["result"][0]

        return result

    def get_global_services(self) -> dict[str, Any]:
        """
        Get the global services database

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = self.api_get(
            "/pm/config/global/obj/firewall/service/custom",
            timeout=10,
        ).json()["result"][0]

        return result

    def get_global_service_group(self, group: str, scope_member: bool = False) -> dict[str, Any]:
        """
        Get a service group object from the global ADOM

        Args:
            group:        The address group to get
            scope_member: Whether the scope member attribute should be included in the response

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = {}
        url: str = f"/pm/config/global/obj/firewall/service/group/{group}"

        if scope_member:
            params = {"option": ["scope member"]}
            response = self.api_get(url, params)

        else:
            response = self.api_get(url)

        result = response.json()["result"][0]

        return result

    def get_global_service_groups(self) -> dict[str, Any]:
        """
        Get the global network service group database

        Returns:
            FortiManager result item
        """
        result: dict[str, Any] = self.api_get(
            "/pm/config/global/obj/firewall/service/group",
            timeout=10,
        ).json()["result"][0]

        return result

    def get_version(self) -> str:
        """
        Get FortiManager version.

        Returns:
            FortiManager version
        """
        fmg_version: str = ""
        payload = {"method": "get", "params": [{"url": "/sys/status"}]}
        response = self.api("post", payload=payload)

        if response.status_code == 200:
            try:
                match: Optional[re.Match[Any]] = re.search(
                    r"(v\d+\.\d+\.\d+)-", response.json()["result"][0]["data"]["Version"]
                )
                fmg_version = str(match.group(1))  # type: ignore

            except (KeyError, IndexError, AttributeError):
                log.debug("did not find any FortiManager version number in response")

        return fmg_version

    def login(self) -> int:
        """
        Login to the FortiManager.

        If the login to the FortiManager was successful it stores the session key in the object.
        We do not use requests.session as the session key is just a string which is saved directly.

        Returns:
            Status code from the FortiManager login
        """
        status: int = 401

        if self.session_path:
            session_file = Path(self.session_path).expanduser() / f"{self.hostname}.key"
            if session_file.exists():
                log.debug("Loading session key from file '%s'", session_file)

                with session_file.open(encoding="UTF-8") as file:
                    self.session_key = file.read()
                    payload = {
                        "method": "get",
                        "params": [{"url": "/sys/status"}],
                        "session": self.session_key,
                    }
                    response = super().api("post", payload=payload)
                    status = response.status_code

                    if (
                        response.status_code == 200
                        and response.json()["result"][0]["status"]["code"] == 0
                    ):
                        log.debug("Session key is still valid")

                    else:
                        log.debug("Session key is invalid")
                        self.session_key = ""

            else:
                log.debug("Session file '%s' does ot exist", session_file)

        if not self.session_key and self.username and self.password:
            log.debug("Login to '%s'", self.hostname)
            payload = {
                "method": "exec",
                "params": [
                    {
                        "data": {"passwd": self.password, "user": self.username},
                        "url": "/sys/login/user",
                    }
                ],
            }
            response = super().api("post", payload=payload)
            if response.status_code == 200:
                if "session" in response.json():
                    log.debug("store session key")
                    self.session_key = response.json()["session"]

                    if self.session_path:
                        log.debug("Saving session key into file '%s'", session_file)

                        with session_file.open("w", encoding="UTF-8") as file:
                            file.write(self.session_key)

            status = response.status_code

        return status

    def logout(self) -> int:
        """
        Logout from FortiManager.

        Returns:
            Status code from the FortiManager logout
        """
        payload: Dict[str, Any] = {
            "method": "exec",
            "params": [{"url": "/sys/logout"}],
            "session": self.session_key,
        }
        response = self.api("post", payload=payload)
        self.session_key = ""
        return response.status_code

    def post(self, adom: str, payloads: Any) -> List[str]:
        """
        POST method to FortiManager.

        You can pass a single payload (Dict) or a list of payloads (List of Dict).

        Args:
            adom:     The ADOM name to issue the set commands to. If you wish to update the
                      Global ADOM specify 'global' as ADOM.
            payloads: One payload (Dict) or a list of payloads (List of Dict)

        Returns:
            Amount of errors ocurred during the set command
        """
        # if payload is a dict convert it to a list with one dict in it.
        if isinstance(payloads, dict):
            payloads = [payloads]

        results = []

        for payload in payloads:
            for i, _ in enumerate(payload["params"]):
                # Here we have to replace the {adom} in the URL entry:
                #    for Global: /pm/config/ global      /obj/firewall/address/{address}
                #    for ADOM:   /pm/config/ adom/{adom} /obj/firewall/address/{address}
                #                            \_________/ --> this is the {adom} part in the payload
                adom_str = "global" if adom.lower() == "global" else f"adom/{adom}"
                payload["params"][i]["url"] = payload["params"][i]["url"].replace(
                    "{adom}", adom_str
                )

            response = self.api("post", payload=payload, timeout=10)
            for result in response.json()["result"]:
                if result["status"]["code"] != 0:
                    log.error("%s: %s", result["status"]["message"], result["url"])
                    results.append(
                        f"{result['status']['message']}: "
                        f"{result['url']} "
                        f"(code: {result['status']['code']})"
                    )

        return results

    def wait_for_task(self, task_id: int, timeout: int = 60) -> List[Any]:
        """
        Wait for a task with a given id for its end and returns the message(s).

        Args:
            id:      Task id to wait for
            timeout: Timeout in seconds

        Returns:
            Message list
        """
        log.debug("Waiting for task id '%s'", task_id)
        messages: List[Any] = []
        payload = {
            "method": "get",
            "params": [{"url": f"/task/task/{task_id}"}],
        }
        percent_cache = -1

        while timeout:
            response = self.api("post", payload=payload)
            percent = response.json()["result"][0]["data"]["percent"]
            if percent > percent_cache:
                log.debug("FortiManager task progress: '%s%%'", percent)
                percent_cache = percent

            if response.status_code == 200:
                if percent >= 100:
                    break

            sleep(1)
            timeout -= 1

        payload = {
            "method": "get",
            "params": [{"url": f"/task/task/{task_id}/line"}],
        }
        response = self.api("post", payload=payload)

        if response.status_code == 200:
            messages = response.json()["result"][0]["data"]
            # enrich the message(s) with the task_id (otherwise it will be lost)
            for i, _ in enumerate(messages):
                messages[i]["task_id"] = task_id

        return messages



"""
The config helper is used to load and set basic configuration for fotoobo. Some configuration
options may be set by the global configuration file which by default points to fotoobo.yaml in the
local directory. Some configuration options may be set by command line parameters as described in
the cli help system.
If there are configuration options available in the global configuration file and also as a command
line option, the command line option takes precedence over the global configuration file option.
"""

import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional

from fotoobo.exceptions import GeneralError
from fotoobo.helpers.files import load_yaml_file


@dataclass(eq=False, order=False)
class Config:
    """
    This is the configuration dataclass for the global configuration options.
    First all the configuration options must be initialized.
    If an option is not defined here it is not guaranteed that it may be used later in the code.
    """

    # set default values
    inventory_file: Path = Path("inventory.yaml")
    logging: Optional[Dict[str, Any]] = None
    audit_logging: Optional[Dict[str, Any]] = None
    no_logo: bool = False
    cli_info: Dict[str, Any] = field(default_factory=dict)
    vault: Dict[str, str] = field(default_factory=dict)

    def load_configuration(  # pylint: disable=too-many-branches
        self, config_file: Optional[Path] = None
    ) -> None:
        """
        Load the global fotoobo configuration file. If the configuration file is not present it just
        will be ignored and all the options remain as they are.
        If a file is present it looks for configuration options to set. Keep in mind that it is not
        mandatory to define every option in the configuration file. For that case make sure that an
        option not present in the configuration file is set with its default.

        Args:
            config_file: fotoobo configuration file to load configuration from
        """

        # If no explicit config file has been given, search at the predefined places
        if config_file is None:
            fotoobo_yaml_in_dot_config = Path("~").expanduser() / ".config" / "fotoobo.yaml"

            # First we check whether there is a project fotoobo.yaml in the
            # current working directory
            if Path("fotoobo.yaml").is_file():
                config_file = Path("fotoobo.yaml")

            # Second we check whether there is a user wide fotoobo.yaml in the .config folder
            elif fotoobo_yaml_in_dot_config.is_file():
                config_file = fotoobo_yaml_in_dot_config

            # If no config file can be found we rest with the defaults
            else:
                return

        if config_file:
            if loaded_config := load_yaml_file(config_file):
                # We need a dict here
                loaded_config = dict(loaded_config)

                # then set the config options from file if set in file
                self.inventory_file = Path(
                    loaded_config.get("inventory", self.inventory_file)
                ).expanduser()

                if not self.inventory_file.is_absolute():
                    self.inventory_file = config_file.parent / self.inventory_file

                self.logging = loaded_config.get("logging", {})
                if not isinstance(self.logging, dict):
                    raise GeneralError("Setting logging has to be a dictionary")
                if self.logging and "log_file" in self.logging:
                    if not isinstance(self.logging["log_file"], dict):
                        raise GeneralError("Setting logging.log_file has to be a dictionary")
                    if not self.logging["log_file"].get("name", ""):
                        raise GeneralError("Missing logging.log_file configuration: name")

                self.audit_logging = loaded_config.get("audit_logging", {})
                if not isinstance(self.audit_logging, dict):
                    raise GeneralError("Setting audit_logging has to be a dictionary")
                if self.audit_logging and "log_file" in self.audit_logging:
                    if not isinstance(self.audit_logging["log_file"], dict):
                        raise GeneralError("Setting audit_logging.log_file has to be a dictionary")
                    if not self.audit_logging["log_file"].get("name", ""):
                        raise GeneralError("Missing audit_logging.log_file configuration: name")

                self.no_logo = loaded_config.get("no_logo", self.no_logo)

                self.vault = loaded_config.get("vault", {})
                if self.vault:
                    # role_id and secret_id may be stored in environment variables (they overwrite
                    # the ones configured in the fotoobo.yaml)
                    if os.getenv("FOTOOBO_VAULT_ROLE_ID"):
                        self.vault["role_id"] = os.getenv("FOTOOBO_VAULT_ROLE_ID", "")

                    if os.getenv("FOTOOBO_VAULT_SECRET_ID"):
                        self.vault["secret_id"] = os.getenv("FOTOOBO_VAULT_SECRET_ID", "")

                    # Check if all the mandatory settings in vault section are given by subtracting
                    # the list of keys in vault config from list of mandatory keys.
                    if (
                        missing := [
                            "url",
                            "namespace",
                            "data_path",
                            "role_id",
                            "secret_id",
                        ]
                        - self.vault.keys()
                    ):
                        raise GeneralError(f"Missing vault configuration: {missing}")


config = Config()



"""
This is the main cli module. If fotoobo is started from the command line the main script starts
app() from within here. Every level of the command line tool is represented by a subdirectory.
If you extend the cli make sure you are in the correct directory or create a new one if there are
more than a few subcommands.

Caution: Use docstrings with care as they are used to print help texts on any command.
"""

# pylint: disable=anomalous-backslash-in-string
import logging
import os
import sys
from pathlib import Path
from typing import Optional

import typer

from fotoobo import tools
from fotoobo.cli import convert, get
from fotoobo.cli.ems import main as ems
from fotoobo.cli.faz import main as faz
from fotoobo.cli.fgt import main as fgt
from fotoobo.cli.fmg import main as fmg
from fotoobo.helpers import cli_path
from fotoobo.helpers.config import config
from fotoobo.helpers.log import Log
from fotoobo.helpers.output import print_logo

app = typer.Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    rich_markup_mode="rich",
)
log = logging.getLogger("fotoobo")


def version_callback(value: bool) -> None:
    """The version callback"""
    if value:
        if not config.no_logo:
            print_logo()

        get.version(False)
        raise typer.Exit()


@app.callback()
def callback(  # pylint: disable=too-many-arguments
    context: typer.Context,
    config_file: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        help="Set the fotoobo configuration file. \[default: fotoobo.yaml]",
        show_default=False,
        metavar="[file]",
    ),
    log_level: str = typer.Option(
        None,
        "--loglevel",
        help="Set the log level. Choose from CRITICAL, ERROR, WARNING, INFO, DEBUG.",
        metavar="[level]",
        show_default=False,
    ),
    nologo: bool = typer.Option(None, "--nologo", help="Do not print the beautiful fotoobo logo."),
    log_quiet: bool = typer.Option(
        None, "--quiet", "-q", help="Disable console logging. \[default: not quiet]"
    ),
    version: Optional[bool] = typer.Option(  # pylint: disable=unused-argument
        None, "--version", "-V", help="Print the fotoobo version.", callback=version_callback
    ),
) -> None:
    """
    The Fortinet Toolbox (fotoobo) - make IT easy

    ─────────────────────────────────────────────

    This is fotoobo, the mighty [bold]Fo[/bold]rtinet [bold]too[/bold]l[bold]bo[/bold]x for
    managing your Fortinet environment. It is meant to be extendable to your needs.

    For detailed documentation see https://fotoobo.readthedocs.io.
    """
    config.load_configuration(config_file)
    config.no_logo = True if nologo else config.no_logo

    if log_level:
        log_level = log_level.upper()

    Log.configure_logging(log_quiet, log_level)

    if not config.no_logo:
        print_logo()

    log.debug("Let the magic begin")

    # Log all the configuration options
    # If a config option is a data-structure (dict) unpack it and write it line-by-line
    # For security reasons all sensitive values are shortened
    for attr in dir(config):
        if attr.startswith("_") or attr in ["config", "load_configuration"]:
            continue

        if attr in ["audit_logging", "logging", "vault"] and getattr(config, attr):
            for sub_attr, value in getattr(config, attr).items():
                if attr == "vault" and sub_attr in ["role_id", "secret_id"]:
                    value = f"{value[:4]}...{value[-4:]}"

                log.debug("Option '%s.%s' is '%s'", attr, sub_attr, value)

            continue

        log.debug("Option '%s' is '%s'", attr, getattr(config, attr))

    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)
    log.audit(f'command="{" ".join(sys.argv)}"')  # type: ignore
    config.cli_info = context.to_info_dict()


@app.command(hidden=True)
def greet(
    name: Optional[str] = typer.Argument(
        None, help="The name of the person to greet.", show_default=False, metavar="[name]"
    ),
    bye: bool = typer.Option(False, "--bye", "-b", help='Also write "bye" at the end.'),
    log_enabled: bool = typer.Option(False, "--log", "-l", help="Enable logging."),
) -> None:
    """
    This is the hidden greeting function.
    It allows you to greet someone with different colors in different languages.
    """
    if not name:
        try:
            name = os.getlogin().capitalize()

        except OSError:  # We need this, will fail on GitHub otherwise...
            name = ""

    tools.greet(str(name), bye, log_enabled)


# fotoobo specific commands
app.add_typer(convert.app, name="convert", help="Convert commands for fotoobo.")
app.add_typer(get.app, name="get", help="Get information about fotoobo or your configuration.")

# commands for the Fortinet products
app.add_typer(ems.app, name="ems", help="Commands for FortiClient EMS.")
app.add_typer(faz.app, name="faz", help="Commands for FortiAnalyzer.")
app.add_typer(fgt.app, name="fgt", help="Commands for FortiGate.")
app.add_typer(fmg.app, name="fmg", help="Commands for FortiManager.")



"""
FortiGate backup utility
"""

import concurrent.futures
import json
import logging
from typing import Optional, Tuple

from rich.progress import Progress

from fotoobo.exceptions import APIError, GeneralError
from fotoobo.fortinet.fortigate import FortiGate
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def backup(
    host: Optional[str] = None,
) -> Result[str]:
    """
    Create a FortiGate configuration backup into a file and optionally upload it to an FTP server.

    Args:
        host: The host from the inventory to get the backup. If no host is given all FortiGate
              devices in the inventory are backed up.

    Returns:
        The Result object with all the results
    """
    result = Result[str]()
    inventory = Inventory(config.inventory_file)
    fgts = inventory.get(host, "fortigate")

    def _get_single_backup(name: str, fgt: FortiGate) -> Tuple[str, str]:
        """Get the configuration backup from a single FortiGate.

        This private method is used for multithreading. It only queries one single FortiGate for its
        configuration backup and returns it.

        Args:
            name: The name of the FortiGate (as defined in the inventory)
            fgt:  The FortiGate object to query

        Returns:
            name: The name of the FortiGate (as defined in the inventory)
            data: The configuration backup of the FortiGate (fgt)
        """
        log.debug("Backup FortiGate '%s'", name)
        data: str = ""

        try:
            data = fgt.backup()
            if data.startswith("#config-version"):
                message = f"Config backup for '{name}' succeeded"
                log.info(message)
                result.push_message(name, message)

            else:
                data_json = json.loads(data)
                message = f"Backup '{name}' failed with error '{data_json['http_status']}'"
                log.error(message)
                result.push_message(name, message, level="error")

        except GeneralError as err:
            result.push_message(name, err.message, level="error")

        except APIError as err:
            result.push_message(name, f"{name} returned {err.message}", level="error")

        return name, data

    with Progress() as progress:
        task = progress.add_task("Getting FortiGate versions...", total=len(fgts))
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            futures = []
            for name, fgt in fgts.items():
                futures.append(executor.submit(_get_single_backup, name, fgt))

            for future in concurrent.futures.as_completed(futures):
                name, configuration_backup = future.result()
                result.push_result(name, configuration_backup)
                progress.update(task, advance=1)

    return result



"""
FortiManager get ADOMs utility
"""

import logging
from typing import Any, Dict, List, Optional, Union

from fotoobo.exceptions.exceptions import GeneralError
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def adoms(host: str) -> Result[str]:
    """
    FortiManager get ADOMs

    Args:
        host: The host from the inventory to get the ADOMs list from

    Returns:
        List of ADOMs where ADOM is a dict with keys: name, version
    """
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    result = Result[str]()
    log.debug("FortiManager get adoms ...")
    fmg.login()
    fmg_adoms = fmg.get_adoms()

    for adom in fmg_adoms:
        result.push_result(adom["name"], f"{adom['os_ver']}.{adom['mr']}")

    fmg.logout()
    return result


def devices(host: str) -> Result[Dict[str, Union[str, List[str]]]]:
    """
    FortiManager get logical devices
    In a cluster only the cluster device is returned, not the physical nodes

    Args:
        host: The FortiManager from the inventory to get the Fortinet devices list from

    Returns:
        List of devices
    """
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    log.debug("FortiManager get devices ...")
    fmg.login()
    payload = {
        "method": "get",
        "params": [
            {
                "loadsub": 1,
                "sortings": [{"build": 1}],
                "option": "object member",
                "url": "/dvmdb/device",
            }
        ],
    }
    response = fmg.api("post", payload=payload)
    fmg.logout()
    result = Result[Dict[str, Union[str, List[str]]]]()

    for device in response.json()["result"][0]["data"]:
        data = {
            "version": f"{device['os_ver']}.{device['mr']}.{device['patch']}",
            "ha_mode": str(device["ha_mode"]),
            "platform": device["platform_str"],
            "desc": device["desc"],
            "ha_nodes": [],
        }
        if device["ha_mode"] == 1:
            ha_nodes = [node["name"] for node in device["ha_slave"]]
            data["ha_nodes"] = ha_nodes

        result.push_result(device["name"], data)

    return result


def policy(
    host: str, adom: str, policy_name: str, fields: Optional[List[str]] = None
) -> Result[List[Dict[str, Any]]]:
    """
    FortiManager get policy

    Args:
        host:
        adom:
        policy_name:
        fields:

    Returns:
        Result
    """
    fields = fields or [
        "status",
        # "_last_hit",  # data-format not clear
        "global-label",
        # "_hitcount",  # data-format not clear (it's not equal to the value in FortiManager)
        "policyid",
        "srcaddr",
        "groups",
        "dstaddr",
        "service",
        "action",
        "send-deny-packet",
        "comments",
    ]
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    log.debug("FortiManager get policy '%s' from '%s' ...", policy_name, adom)
    fmg.login()
    payload = {
        "method": "get",
        "params": [
            {
                "option": "object member",
                "url": f"/pm/config/adom/{adom}/pkg/{policy_name}/firewall/policy",
            }
        ],
        "session": "",
        "id": 1,
    }
    result = fmg.api("post", payload=payload, timeout=30)
    data = result.json()

    if data["result"][0]["status"]["code"] != 0:
        code = data["result"][0]["status"]["code"]
        message = data["result"][0]["status"]["message"]
        log.error("FortiManager '%s' returned '%s': '%s'", host, code, message)
        raise GeneralError(f"FortiManager {host} returned {code}: {message}")

    policies = []
    for pol in data["result"][0]["data"]:
        policies.append({field: pol.get(field, None) for field in fields})

    fmg.logout()
    out_result = Result[List[Dict[str, Any]]]()
    out_result.push_result(host, policies)
    return out_result


def version(host: str) -> Result[str]:
    """
    FortiManager get version

    Args:
        host: The host from the inventory to get the version

    Returns:
        Version data in a dict with keys: host, version
    """
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    log.debug("FortiManager get version ...")
    fmg_version = fmg.get_version()
    result = Result[str]()
    result.push_result(host, fmg_version)
    return result



"""
The FortiGate commands
"""

import logging
import os
from datetime import datetime
from pathlib import Path

import typer

from fotoobo import tools
from fotoobo.cli.fgt import config_commands as config
from fotoobo.cli.fgt import get_commands as get
from fotoobo.cli.fgt import monitor_commands as monitor
from fotoobo.exceptions import GeneralWarning
from fotoobo.helpers import cli_path
from fotoobo.helpers.config import config as fotoobo_config
from fotoobo.helpers.files import create_dir, file_to_ftp, file_to_zip
from fotoobo.inventory import Inventory

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fgt subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


app.add_typer(get.app, name="get", help="FortiGate get commands.")
app.add_typer(monitor.app, name="monitor", help="FortiGate monitor commands.")
app.add_typer(config.app, name="config", help="FortiGate config file commands.")


@app.command()
def backup(
    host: str = typer.Argument(
        "",
        help="The FortiGate to backup (must be defined in the inventory). "
        "Backups all if left empty.",
        show_default=False,
        metavar="[host]",
    ),
    backup_dir: Path = typer.Option(
        None,
        "--backup-dir",
        "-b",
        help="The directory to save the backup(s) to. Default is the current working directory.",
        show_default=False,
        metavar="backup_dir",
    ),
    ftp_server: str = typer.Option(
        None,
        "--ftp",
        "-f",
        help="The ftp configuration from the inventory to send the backup to.",
        metavar="server",
    ),
    smtp_server: str = typer.Option(
        None,
        "--smtp",
        "-s",
        help="The smtp configuration from the inventory to send potential errors to.",
        metavar="server",
    ),
) -> None:
    """
    Backup one or more FortiGate(s).
    """
    inventory = Inventory(fotoobo_config.inventory_file)

    if not backup_dir:
        backup_dir = Path.cwd()

    create_dir(backup_dir)
    result = tools.fgt.backup(host)

    for name, data in result.all_results().items():
        config_file = backup_dir / Path(name).with_suffix(".conf")

        if config_file.is_file():
            os.remove(config_file)

        config_file.write_text(data, encoding="UTF-8")

        if not config_file.is_file():
            result.push_message(name, f"backup file for '{name}' does not exist")
            continue

        if ftp_server:
            if ftp_server in inventory.assets:
                server = inventory.assets[ftp_server]
                log.debug("Compressing configuration '%s'", name)
                time: str = datetime.now().strftime("%Y%m%d-%H%M")
                zip_file = backup_dir / Path(name + "-" + time + ".conf.zip")
                file_to_zip(config_file, zip_file)
                file_to_ftp(zip_file, server)
                os.remove(zip_file)
                result.push_message(name, f"Uploaded config file for '{name}' to '{ftp_server}'")

            else:
                raise GeneralWarning(f"FTP server '{ftp_server}' not found in inventory")

    if smtp_server and smtp_server in inventory.assets:
        result.send_messages_as_mail(inventory.assets[smtp_server], "error")



"""
A module for defining a generic device
"""

from typing import Union


class GenericDevice:
    """
    This is the class for a generic device.

    All kwarg will be unpacked and written to the instance as attributes. There are some mandatory
    attributes which will be filled with a default value.
    """

    def __init__(self, **kwargs: Union[str, int, bool]) -> None:
        """init"""
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.hostname = getattr(self, "hostname", "")
        self.type = getattr(self, "type", "generic")



"""
FortiManager assign utility
"""

import logging
from pathlib import Path

from fotoobo.exceptions.exceptions import GeneralWarning
from fotoobo.helpers.config import config
from fotoobo.helpers.files import load_json_file
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def assign(adoms: str, policy: str, host: str, timeout: int = 60) -> Result[str]:
    """
    Assign the global policy to the given ADOM

    Args:
        adoms:  The ADOMs to assign the global policy to. Specify multiple ADOMs as a comma
                separated list (no spaces).
        policy: Specify the global policy to assign [Default: 'default'].
        host:   The FortiManager defined in inventory.
        timeout Timeout in sec. to wait for the FortiManager task to finish [Default: 60].
    """
    result = Result[str]()
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    log.debug("Assigning global policy/objects to ADOM '%s'", adoms)

    task_id = fmg.assign_all_objects(adoms=adoms, policy=policy)
    if task_id > 0:
        log.info("Created FortiManager task id '%s'", task_id)
        messages = fmg.wait_for_task(task_id, timeout=timeout)

        for message in messages:
            level = "debug" if message["state"] == 4 else "error"
            elapsed = (
                str(message["end_tm"] - message["start_tm"]) + " sec"
                if message["end_tm"] > 0
                else "unfinished"
            )
            result_message = f"{message['task_id']}: {message['name']}"

            if message["detail"]:
                result_message += f" / {message['detail']}"

            result_message += f" ({elapsed})"
            getattr(log, level)(result_message)
            result.push_message(host, result_message, level)

            if message["history"]:
                for line in message["history"]:
                    result_message = f"- {line['detail']}"
                    getattr(log, level)(result_message)
                    result.push_message(host, result_message, level)

    return result


def post(file: Path, adom: str, host: str) -> Result[str]:
    """
    POST the given configuration from a JSON file to the FortiManager

    Args:
        file: The configuration file to add the configuration from
        adom: The ADOM to assign the global policy to
        host: The FortiManager defined in inventory

    Returns:
        Result

    Raises:
        GeneralWarning
    """
    if not (payloads := load_json_file(file)):
        raise GeneralWarning(f"There is no data in the given file ({file})")

    inventory = Inventory(config.inventory_file)
    result: Result[str] = Result()
    fmg = inventory.get_item(host, "fortimanager")
    log.debug("FortiManager post command ...")
    log.info("Start posting assets to '%s'", host + "/" + adom)

    result_list = fmg.post(adom, payloads)
    if result_list:
        for line in result_list:
            result.push_message(host, line, "error")

    return result



"""
The FortiGateInfo class
"""

from dataclasses import dataclass


@dataclass(eq=False, order=False)
class FortiGateInfo:
    """
    This dataclass holds FortiGate meta information from the configuration file.
    """

    # pylint: disable=too-many-instance-attributes

    # attributes in alphabetical order
    buildno: str = ""
    conf_file_ver: str = ""
    global_vdom: str = ""
    hostname: str = ""
    model: str = ""
    opmode: str = ""
    os_version: str = ""
    type: str = ""
    user: str = ""
    vdom: str = ""  # 0 in single vdom mode, 1 in multiple vdom mode



"""
The FortiGate configuration class represents the whole or parts of a FortiGate configuration
"""

import logging
from pathlib import Path
from typing import IO, Any, Dict, List, Optional

from fotoobo.exceptions import GeneralWarning
from fotoobo.helpers.files import load_json_file, save_json_file

from .fortigate_info import FortiGateInfo

log = logging.getLogger("fotoobo")


class FortiGateConfig:
    """
    The FortiGateConfig class represents a FortiGate configuration (or parts of it)
    """

    _config_path: List[str] = []

    def __init__(
        self,
        global_config: Optional[Dict[str, Any]] = None,
        vdom_config: Optional[Dict[str, Any]] = None,
        info: Optional[Dict[str, str]] = None,
    ) -> None:
        self.global_config = global_config or {}
        self.vdom_config = vdom_config or {}
        self.info = FortiGateInfo(**(info or {}))
        try:
            self.info.hostname = self.global_config["system"]["global"]["hostname"]

        except KeyError:
            self.info.hostname = "HOSTNAME UNKNOWN"

    def get_configuration(self, scope: str = "global", path: str = "/") -> Any:
        """
        Return a config snippet from the given configuration path.

        Args:
            scope: The configuration part to get the configuration from (global|vdom)
            path:  From where should the configuration come?
            Examples:
            /                           : The whole configuration (default)
            /system/global/...          : Sub-path in global configuration
            /system/global/admin-scp    : Exact configuration option
            /root/firewall/address/...  : Sub-path in root vdom

        Returns:
            FortiGateConfig configuration value or snippet from the given path.
            If the path or value does not exist None is returned.
        """
        log.debug("Getting configuration from scope: '%s'", scope)
        log.debug("Getting configuration from path: '%s'", path)
        # split the path by / and remove all empty parts
        path_list: List[str] = [p for p in path.strip("/").split("/") if p]
        config: Any = {}

        if scope == "global":
            config = self.global_config

        elif scope == "vdom":
            config = self.vdom_config

        if len(path_list) > 0:
            for key in path_list:
                if key in config:
                    config = config[key]

                else:
                    config = {}
                    break

        return config

    def get_vdoms(self) -> List[str]:
        """
        Get the list of configured VDOMs.
        If it is a single vdom configuration an empty list will be returned.

        Returns:
            List of VDOMs
        """
        vdoms: List[str] = []
        if self.info.vdom == "1":
            vdoms = list(self.vdom_config.keys())

        return vdoms

    def save_configuration_file(self, configuration_file: Path) -> None:
        """
        Save the configuration to a json configuration file.

        Args:
            configuration_file: The json file to load the FortiGate configuration from
        """
        save_json_file(
            configuration_file,
            {"global": self.global_config, "vdom": self.vdom_config, "info": self.info.__dict__},
        )

    @staticmethod
    def parse_configuration_file(configuration_file: Path) -> "FortiGateConfig":
        """
        Parse the FortiGate configuration from a file into a python object

        Args:
            configuration_file: The filename of the FortiGate configuration file

        Returns:
            The parsed FortiGate configuration object
        """
        log.debug("Start configuration parser with file '%s'", configuration_file)

        FortiGateConfig._config_path = []
        with configuration_file.open(encoding="UTF-8") as forti_file:
            parsed_config = FortiGateConfig._parse_to_dict(forti_file)

        global_config: Dict[str, Any] = {}
        vdom_config: Dict[str, Any] = {}

        if "info" not in parsed_config:
            raise GeneralWarning(f"There is no info in {configuration_file}")

        info = parsed_config["info"]
        parsed_config.pop("info")

        if info["vdom"] == "0":
            global_config["system"] = {}
            global_config["system"] = parsed_config["system"]
            parsed_config.pop("system")
            vdom_config["root"] = {}
            vdom_config["root"] = parsed_config

        else:
            global_config = parsed_config["global"]
            vdom_config = parsed_config["vdom"]

        return FortiGateConfig(global_config, vdom_config, info)

    @staticmethod
    def load_configuration_file(configuration_file: Path) -> "FortiGateConfig":
        """
        Create a FortiGateConfig object from a json configuration file.

        Args:
            configuration_file: The json file to load the FortiGate configuration from

        Returns:
            FortiGate configuration object
        """
        data = load_json_file(configuration_file)
        return FortiGateConfig(data["global"], data["vdom"], data["info"])  # type: ignore

    @staticmethod
    def _config_convert_dict_to_list(config: Dict[str, Any]) -> List[Any]:
        """
        Convert a FortiGate configuration list like part to a configuration list.

        Args:
            config: A FortiGate configuration list like part from a dict

        Returns:
            Configuration list as list
        """
        configs: List[str] = []
        for key, value in config.items():
            new_config = value
            new_config["id"] = int(key)
            configs.append(new_config)

        return configs

    @staticmethod
    def _config_is_list(config: Dict[str, Any]) -> bool:
        """
        Check if a FortiGate configuration part contains a list of elements.

        Args:
            config: The FortiGate configuration part to check

        Returns:
            Is it a list (True) or not (False)
        """
        is_list = True
        if len(config) == 0:
            is_list = False

        for key in config.keys():
            if not key.isnumeric():
                is_list = False

        return is_list

    @staticmethod
    def _get_nested_dict(configs: List[str], new_data: Any) -> Dict[str, Any]:
        """
        Create a nested configuration dict from a list of strings and add new data to the last key.

        Args:
            configs:  Configuration leafs to process
            new_data: Data to add to the last key

        Returns:
            The nested configuration
        """
        out_dict = {}
        key = configs.pop(0)
        if len(configs) > 0:
            out_dict[key] = FortiGateConfig._get_nested_dict(configs, new_data)

        else:
            out_dict[key] = new_data

        return dict(out_dict)

    @staticmethod
    def _parse_config_comment(info: Dict[str, Any], line: str) -> Dict[str, str]:
        """
        Parses a comment line in a FortiGate configuration and extracts interesting values.
        The values are then written to an info dict as meta-information.

        Args:
            info: The meta information dict to add information to
            line: The FortiGate configuration comment line to parse

        Returns:
            The updated meta information dict
        """
        line = line.lstrip("#").strip()
        if line.startswith("config-version"):
            value = line.split("=", maxsplit=1)[1]
            info["model"], info["os_version"], info["type"] = value.split("-")[0:3]
            for part in line.split(":")[1:]:
                info[part.split("=")[0]] = part.split("=")[1]

        elif line.count("=") == 1:
            info[line.split("=")[0]] = line.split("=")[1]

        return info

    @staticmethod
    # pylint: disable=too-many-branches
    def _parse_to_dict(config_file: IO[str]) -> Any:
        # should be Union[Dict[str, Any], List[Any]]
        """
        Fabric function to create a FortiGateConfig object from a backup configuration file
        This recursive method parses a FortiGate configuration from a file line by line

        Args:
            config_file: FortiGate configuration file object

        Returns:
            A dict which contains the parsed FortiGate configuration
        """
        config: Any = {}
        info: Dict[str, str] = {}
        multiline: str = ""
        multiline_key: str = ""

        for line in config_file:
            line = line.strip()

            # handle empty lines
            if not line:
                continue

            # handle comment lines
            if line.startswith("#"):
                info = FortiGateConfig._parse_config_comment(info, line)

            # handle multiline strings (do that before all the other logic)
            if multiline:
                multiline += f"\n{line}"
                if line.endswith('"'):
                    config[multiline_key], multiline = multiline.strip('"'), ""

                continue

            # check if a multiline string starts. begin with "set" and uneven amount of quotes (")
            if line.startswith("set ") and line.count('"') % 2 == 1 and not line.endswith('"'):
                _, multiline_key, value = line.split(maxsplit=2)  # first part is always "set"
                multiline += value
                continue

            # handle configuration option
            if line.startswith("set "):
                _, key, value = line.split(maxsplit=2)  # first part ist always "set"
                config[key] = " ".join(value.replace('"', "").split())

            # handle recursion
            if line.startswith("config vdom") and len(FortiGateConfig._config_path) == 1:
                continue

            if line.startswith("config "):
                if len(line[7:].split(" ")) == 1:
                    FortiGateConfig._config_path.append(line[7:])
                    config[line[7:].strip('"')] = FortiGateConfig._parse_to_dict(config_file)

                else:
                    FortiGateConfig._config_path.append(line[7:])
                    config_path = [word.strip('"') for word in line[7:].split()]
                    temp_config = FortiGateConfig._get_nested_dict(
                        config_path[1:], FortiGateConfig._parse_to_dict(config_file)
                    )

                    if config_path[0] not in config:
                        config[config_path[0]] = {}

                    config[config_path[0]] = {**config[config_path[0]], **temp_config}

            if line.startswith("edit "):
                FortiGateConfig._config_path.append(line[5:])
                config[line[5:].strip('"')] = FortiGateConfig._parse_to_dict(config_file)

            # handle section ends
            if line == "end":
                FortiGateConfig._config_path.pop()
                if FortiGateConfig._config_is_list(config):
                    config = FortiGateConfig._config_convert_dict_to_list(config)

                return config

            if line == "next":
                FortiGateConfig._config_path.pop()
                return config

        # append info dict to config if it's set
        if len(info) > 0:
            config["info"] = info

        return config
